#!/usr/bin/env Rscript

# Read me ----
# Mosaicing the rasters created in "ConversionScript"
# This is for Business-as-usual, change the scenario on L29, 136, 144

# Packages ---
# Keep these to a bare minimum as they take up space
library(raster)
library(data.table)
library(doMC)
rasterOptions(maxmemory = 1e+11)

# Getting 2010 pasture ----
pasture_2010 <- raster("Data/Pasture2010.Mollweide.1.5km_Proportion_griddedlivestock_harmonised.tif")
pasture_2010_vector <- cbind(c(getValues(pasture_2010)))

# Getting row index vectors [DON'T CHANGE THESE FOR EACH SCENARIO] -----
# Rasters for the row indices of individual countries
row_index_rasters <- list.files(path = "Data/Impact_Fine_Country", 
                                pattern = "Index",
                                full.names = TRUE)

# Get a global raster of row indices
country_id_raster_global <- raster("Data/MollweideCountryID_1.5km.tif")

# Vector of file names for changed country rasters ------
# Just vary thesefiles for different scenarios
pasture_rasters_long <- list.files("Outputs/BAU_EATLancet/tifs", 
                           pattern = "pasture",
                           full.names = TRUE)
# Limiting to mean rasters (vs. standard deviation)
pasture_rasters_long_mean <- pasture_rasters_long[grep('mean', pasture_rasters_long)]
pasture_rasters_long_sd <- pasture_rasters_long[grep('sd', pasture_rasters_long)]

# Getting years
yrs <- paste(c(seq(2010, 2045, by = 5)),
             c(seq(2015, 2050, by = 5)),
             sep = "_")

world_dt <- data.table(
  cell = 1:289285431, # Number of cells in global raster
  pasture_orig = rep(-1, 289285431))

world_dt$pasture_new <- world_dt$pasture_orig

setkey(world_dt, cell) 
 # Function to run -----
raster_fun <- function(raster_list, orig_raster, orig_values, id_raster, year, dt, directory){
  pasture_rasters <- raster_list[grep(year, raster_list)]
  mean_sd <- if(grepl("mean", pasture_rasters[1])) "mean" else "sd"
  
  dt_copy <- copy(world_dt) # Create a copy of the data.table, so we don't have to re-make it
  
  # Cycle through the rasters...
  for(i in 1:length(pasture_rasters)){
    print(i) # Print progress
    # Create a data.table with the proportion of pastureland and the row indexes.
    tmp_dt <- data.table(pasture = getValues(raster(pasture_rasters[[i]])),
                         row_index = getValues(raster(row_index_rasters[grep(substring(gsub("^.*/tifs/", "", 
                                                                                            pasture_rasters[[i]]), 
                                                                                       first = 1, last =3),
                                                                             row_index_rasters)])))
    # Hopefully not needed, but better safe than sorry
    tmp_dt <- tmp_dt[order(tmp_dt$row_index)]

    dt_copy[tmp_dt[!is.na(tmp_dt$pasture), row_index], 
            # Returns the row_index of the cells in the temporary data.table 
            #     where the pastureland cell is not NA
            pasture_new := tmp_dt[!is.na(tmp_dt$pasture), pasture]]
    # the ":=" syntax is some madcap binary search thing which is lightning
  }
  # Making raster
  tmp_new <- raster(matrix(dt_copy$pasture_new, 
                           nrow = nrow(id_raster),
                           byrow = FALSE))
  
  # Converting raster back into a matrix
  tmp_new <- cbind(c(getValues(tmp_new)))
  
  tmp_new[tmp_new == -1] <- orig_values[tmp_new == -1]
  
  # And also calculating delta pasture extent
  delta.ag.vector <- tmp_new - orig_values
  
  # And converting back into a raster
  # First absolute pasture extent
  tmp_new <- raster(matrix(tmp_new,
                           nrow = nrow(id_raster),
                           byrow = TRUE))
  crs(tmp_new) <- "+proj=moll +lon_0=0 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs"
  extent(tmp_new) <- extent(orig_raster)
  # Second delta pasture extent
  if(mean_sd == "mean"){
    tmp_new_delta <- raster(matrix(delta.ag.vector,
                                   nrow = nrow(id_raster),
                                   byrow = TRUE))
    crs(tmp_new_delta) <- "+proj=moll +lon_0=0 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs"
    extent(tmp_new_delta) <- extent(orig_raster)
  }
  
  if(mean_sd == "mean"){
    writeRaster(tmp_new,
                paste("Outputs/WorldAgProjections_EATLancet/", directory, "/",
                      "Pasture_",
                      year, ".tif",
                      sep = ""),
                overwrite = TRUE)
    # Second delta extent
    writeRaster(tmp_new_delta,
                paste("Outputs/WorldAgProjections_EATLancet/", directory, "/",
                      "DeltaPasture_",
                      year, ".tif",
                      sep = ""),
                overwrite = TRUE)
  } else {
    writeRaster(tmp_new,
                paste("Outputs/WorldAgProjections_EATLancet/", directory, "/",
                      "Pasture_sd",
                      year, ".tif",
                      sep = ""),
                overwrite = TRUE)
  }
  
}

# Run across multiple cores  -----
registerDoMC(cores = length(yrs))

foreach(j = 1:length(yrs)) %dopar% raster_fun(raster_list = pasture_rasters_long_mean, 
                                              orig_raster = pasture_2010, 
                                              orig_values = pasture_2010_vector,
                                              id_raster = country_id_raster_global, 
                                              year = yrs[j], 
                                              dt = world_dt,
                                              directory = "BAU")

foreach(j = 1:length(yrs)) %dopar% raster_fun(raster_list = pasture_rasters_long_sd, 
                                              orig_raster = pasture_2010, 
                                              orig_values = pasture_2010_vector,
                                              id_raster = country_id_raster_global, 
                                              year = yrs[j], 
                                              dt = world_dt,
                                              directory = "BAU")
